// For examples, see:
// https://thegraybook.vvvv.org/reference/extending/writing-nodes.html#examples


using System.Text;
using System.Buffers.Binary;
using Stride.Core.Mathematics;

namespace Main;

public class SensaCell
{
    // NOTE:
    // convert all [Input] and [Output] spreads to function in/out parameters

    //[Input("Input", IsSingle = true)]
    //IDiffSpread<Stream> FStreamIn;
    //[Output("ID", Visibility = PinVisibility.OnlyInspector)]
    //ISpread<string> FId;
    // ...

    // NOTE:
    // the function is 'pure'
    // meaning it does not have to track any values between subsequent executions (== no state)
    // it just takes an input and converts it
    // therefore it can be a static function - equivalent to VL operation
    // and does not need to be a class member function
    // and also does not need to be a process node


    public static void Decode(
        ReadOnlyMemory<byte> memory,
        out string Id,
        out int ChunkSize,
        out int PlaneCount,
        out int Type,
        out int DimCount,
        out int[] Dim,
        out int[] Dimstride,
        out int DataSize,
        out double Time,
        out int[] Data,
        out Color4[] DataCol
        )
    { 
        // NOTE
        // no need for intermediate variables anymore
        // in beta plugins, there was the spreadmax loop and all pins were spreads
        // we can directly write into output variables
        
        // unused
        // First Chunk: int32 id (4 bytes containing chars), int32 size
        // chunkId1 = memory[0..4].ToArray();
        // chunkSize1 = BinaryPrimitives.ReadInt32LittleEndian(memory[4..8].Span);
        
        // NOTE
        // code uses newer c# features like Ranges[start..end]
        // and Memory<T>, Span<T> and Slice which avoids copying data

        Id = ASCIIEncoding.ASCII.GetString(memory[8..12].Span);

        // NOTE
        // unfortunately .NET BinaryReader doesn't support .ReadInt32(BigEndian) anymore
        // that was only available under .netframework

        ChunkSize = BinaryPrimitives.ReadInt32BigEndian(memory[12..16].Span);
        PlaneCount = BinaryPrimitives.ReadInt32BigEndian(memory[16..20].Span);
        Type = BinaryPrimitives.ReadInt32BigEndian(memory[20..24].Span);
        DimCount = BinaryPrimitives.ReadInt32BigEndian(memory[24..28].Span);

        int[] dim = new int[32];
        for (int i = 0; i < 32; i++)
        {
            var start = 28 + i * 4;
            var end = start + 4;
            dim[i] = BinaryPrimitives.ReadInt32BigEndian(memory[start..end].Span);
        }

        int[] dimstride = new int[32];
        for (int i = 0; i < 32; i++)
        {
            var start = 156 + i * 4;
            var end = start + 4;
            dimstride[i] = BinaryPrimitives.ReadInt32BigEndian(memory[start..end].Span);
        }
        DataSize = BinaryPrimitives.ReadInt32LittleEndian(memory[284..288].Span);
        Time = BinaryPrimitives.ReadInt32LittleEndian(memory[288..292].Span);


        Dim = dim[..DimCount];
        Dimstride = dimstride[..DimCount];

        // process data for output
        var data = memory.Slice(292, DataSize).Span;

        int pixelCount = Dim[0] * Dim[1];
        int dataCount = (int)(DataSize / PlaneCount);
        Data = new int[pixelCount * PlaneCount];
        DataCol = new Color4[pixelCount];

        // how many bytes are used to fill up the block?
        int padding = Dimstride[1] - Dimstride[0] * Dim[0];
        int xBlockSize = Dim[0] * PlaneCount + padding;

        // decode to values
        int count = 0;
        for (int j = 0; j < DataSize; j++)
        {
            // only consider data that is not padding
            if (j % xBlockSize < Dim[0] * PlaneCount)
            {
                Data[count] = data[j];
                count++;
            }
        }

        // decode to color
        if (PlaneCount <= 4)
        {
            for (int j = 0; j < dataCount; j++)
            {
                Color4 col = new Color4();
                // cases emulate behaviour in max
                switch (PlaneCount)
                {
                    case 1:
                        col.R = Data[j * PlaneCount] / 255.0f;
                        col.G = Data[j * PlaneCount] / 255.0f;
                        col.B = Data[j * PlaneCount] / 255.0f;
                        col.A = 1;
                        break;
                    case 2:
                        col.G = Data[j * PlaneCount] / 255.0f;
                        col.R = Data[j * PlaneCount + 1] / 255.0f;
                        col.B = Data[j * PlaneCount + 1] / 255.0f;
                        col.A = 1;
                        break;
                    case 3:
                        col.B = Data[j * PlaneCount] / 255.0f;
                        col.R = Data[j * PlaneCount + 1] / 255.0f;
                        col.G = Data[j * PlaneCount + 2] / 255.0f;
                        col.A = 1;
                        break;
                    case 4:
                        col.A = Data[j * PlaneCount] / 255.0f;
                        col.R = Data[j * PlaneCount + 1] / 255.0f;
                        col.G = Data[j * PlaneCount + 2] / 255.0f;
                        col.B = Data[j * PlaneCount + 3] / 255.0f;
                        break;
                    default:
                        break;
                }
                DataCol[j] = col;
            }
        }

    }
}

